/* This file is part of tagr.
   Copyright (C) 2000, 2005,  Max Bouglacoff, Sergey Poznyakoff

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
   02111-1307, USA. */

#ifdef HAVE_CONFIG_H
# include <config.h>
#endif

#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <ctype.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/wait.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <argp.h>
#include <netdb.h>
#include <signal.h>
#include <errno.h>
#include <string.h>
#include <pwd.h>
#include <grp.h>
#include <tagr.h>
#include <report.h>

#ifndef RETSIGTYPE
# define RETSIGTYPE void
#endif

static RETSIGTYPE sig_quit (int);
static RETSIGTYPE sig_fatal (int);
static RETSIGTYPE sig_hup (int);
static RETSIGTYPE sig_child (int);

char pidfile[] = TAGR_PIDFILE;
char i_recv_buffer[reply_size (MAXADDR)];
char *recv_buffer = i_recv_buffer;
char *configfile = TAGR_CONFIGFILE;
char *html_template = TAGR_TEMPLATE;
char *user;
char *basedir;
int port;
int sockfd;
int foreground = 0;
int single_process_option = 0;
int verbose_level;
static int rebuild_option;

static int import_option;

static int list_option;

static char *check_mode = 0;
static char *user_option = NULL;
static char *html_template_option = NULL;

const char *program_version = "tagr (" PACKAGE_STRING ")";
static char doc[] = "tagr -- traffic analyzer and grapher";
static char args_doc[] = "";

enum {
  OPT_IMPORT = 256
};

static struct argp_option options[] = {
  {"foreground",  'f', NULL, 0, "Run in foreground mode", 0},
  {"config-file", 'c', "FILE", 0, "Read specified configuration FILE", 0},
  {"single-process", 's', NULL, 0, "Run in single-process mode", 0},
  {"html-template", 't', "FILE", 0, "Use given HTML template file", 0},
  {"test", 'T', "DEBUG-FLAGS", 0, "Run in test mode", 0},
  {"user", 'u', "USER-NAME", 0, "Run with this user privileges", 0},
  {"import", OPT_IMPORT, NULL, 0,
   "Import old (mrtg-style) log files from DIR. If DIR is not given, use basedir value from the configuration file", 0 },
  {"rebuild", 'b', NULL, 0, "Rebuild graphs using existing statistics", 0},
  {"verbose", 'v', NULL, 0, "Increase verbosity level", 0},
  {"list",    'l', NULL, 0, "List contents of the rate database", 0},
  {NULL}
};

static error_t
parse_opt (int key, char *arg, struct argp_state *state)
{
  switch (key)
    {
    case 'f':
      foreground++;
      break;

    case 'b':
      rebuild_option = 1;
      break;
      
    case 'c':
      configfile = arg;
      break;

    case 'l':
      list_option = 1;
      break;
	
    case 's':
      single_process_option++;
      break;
	  
    case 't':
      html_template_option = arg;
      break;
	  
    case 'T':
      check_mode = arg;
      break;

    case 'u':
      user_option = arg;
      break;

    case 'v':
      verbose_level++;
      break;
      
    case OPT_IMPORT:
      import_option++;
      break;
      
    default:
      return ARGP_ERR_UNKNOWN;
    }
  return 0;
}

static struct argp argp = {
  options,
  parse_opt,
  args_doc,
  doc,
  NULL,
  NULL,
  NULL
};

    
/* Change to the given uid/gid. Clear the supplementary group list.
   On success returns 0.
   On failure returns 1 (or exits, depending on topt settings. See
   anubis_error) */
static int
change_privs (uid_t uid, gid_t gid)
{
  int rc = 0;
  gid_t emptygidset[1];

  /* Reset group permissions */
  emptygidset[0] = gid ? gid : getegid ();
  if (geteuid () == 0 && setgroups (1, emptygidset))
    {
      logmsg (L_ERR, "setgroups(1, %lu) failed: %s",
	      (u_long) emptygidset[0], strerror (errno));
      rc = 1;
    }

  /* Switch to the user's gid. On some OSes the effective gid must
     be reset first */

#if defined(HAVE_SETEGID)
  if ((rc = setegid (gid)) < 0)
    logmsg (L_ERR, "setegid(%lu) failed: %s", (u_long) gid, strerror (errno));
#elif defined(HAVE_SETREGID)
  if ((rc = setregid (gid, gid)) < 0)
    logmsg (L_ERR, "setregid(%lu,%lu) failed: %s",
	    (u_long) gid, (u_long) gid, strerror (errno));
#elif defined(HAVE_SETRESGID)
  if ((rc = setresgid (gid, gid, gid)) < 0)
    logmsg (L_ERR, "setresgid(%lu,%lu,%lu) failed: %s",
	    (u_long) gid, (u_long) gid, (u_long) gid, strerror (errno));
#endif

  if (rc == 0 && gid != 0)
    {
      if ((rc = setgid (gid)) < 0 && getegid () != gid)
	logmsg (L_ERR, "setgid(%lu) failed: %s",
		(u_long) gid, strerror (errno));
      if (rc == 0 && getegid () != gid)
	{
	  logmsg (L_ERR, "cannot set effective gid to %lu: %s",
		  (u_long) gid, strerror (errno));
	  rc = 1;
	}
    }

  /* now reset uid */
  if (rc == 0 && uid != 0)
    {
      uid_t euid;

      if (setuid (uid)
	  || geteuid () != uid
	  || (getuid () != uid && (geteuid () == 0 || getuid () == 0)))
	{

#if defined(HAVE_SETREUID)
	  if (geteuid () != uid)
	    {
	      if (setreuid (uid, -1) < 0)
		{
		  logmsg (L_ERR,
			  "setreuid(%lu,-1) failed: %s",
			  (u_long) uid, strerror (errno));
		  rc = 1;
		}
	      if (setuid (uid) < 0)
		{
		  logmsg (L_ERR,
			  "second setuid(%lu) failed: %s",
			  (u_long) uid, strerror (errno));
		  rc = 1;
		}
	    }
	  else
#endif
	    {
	      logmsg (L_ERR, "setuid(%lu) failed: %s",
		      (u_long) uid, strerror (errno));
	      rc = 1;
	    }
	}


      euid = geteuid ();
      if (uid != 0 && setuid (0) == 0)
	{
	  logmsg (L_ERR, "seteuid(0) succeeded when it should not");
	  rc = 1;
	}
      else if (uid != euid && setuid (euid) == 0)
	{
	  logmsg (L_ERR, "cannot drop non-root setuid privileges");
	  rc = 1;
	}
    }
  return rc;
}

void
change_user ()
{
  struct passwd *pwd;

  if (user == NULL)
    return;

  pwd = getpwnam (user);
  if (pwd)
    {
      if (change_privs (pwd->pw_uid, pwd->pw_gid))
	exit (1);

      chdir (pwd->pw_dir);
    }
}


#define CHECK_USAGE(cond, opt, mode_opt) \
 if (cond) \
   die (1, "%s is meaningless with %s", opt, mode_opt);

int
main (int argc, char **argv)
{
  int index, rc;
  struct sockaddr salocal, saremote;
  struct sockaddr_in *sin;
  fd_set read_fds;
  int salen;

  argp_program_bug_address = "<" PACKAGE_BUGREPORT ">";
  init_syslog (argv[0]);

  if (argp_parse (&argp, argc, argv, 0, &index, NULL))
    exit (1);

  argc -= index;
  argv += index;

  if (argc != 0 && !(list_option || import_option))
    die (1, "Too many arguments");

  if (check_mode)
    {
      CHECK_USAGE (import_option, "--import", "--test");
      CHECK_USAGE (rebuild_option, "--rebuild", "--test");
      CHECK_USAGE (list_option, "--list", "--test");
      
      for (; *check_mode; check_mode++)
	switch (*check_mode)
	  {
	  case 'c':
	    exit (readconfig ());
	    
	  case 't':
	    exit (check_template ());
	    
	  default:
	    die (1, "Unknown check letter: %c", *check_mode);
	  }
    }
  
  if (readconfig ())
    exit (2);

  if (import_option)
    {
      if (argc)
	while (argc--)
	  import (*argv++);
      else
	import (basedir);
    }

  if (rebuild_option)
    rebuild ();

  if (list_option)
    list_db ();
  
  if (rebuild_option || import_option || list_option)
    exit (0);
  
  if (user_option)
    user = user_option;
  if (html_template_option)
    html_template = html_template_option;
  
  sockfd = socket (AF_INET, SOCK_DGRAM, 0);
  if (sockfd < 0)
    die (1, "socket: %s", strerror (errno));

  sin = (struct sockaddr_in *) &salocal;
  memset (sin, 0, sizeof (salocal));
  sin->sin_family = AF_INET;
  sin->sin_addr.s_addr = INADDR_ANY;
  sin->sin_port = htons (port);

  if (bind (sockfd, &salocal, sizeof (*sin)) < 0)
    die (1, "bind: %s", strerror (errno));
  
  signal (SIGHUP, sig_hup);
  signal (SIGUSR1, SIG_IGN);
  signal (SIGUSR2, SIG_IGN);
  signal (SIGQUIT, sig_quit);
  signal (SIGTERM, sig_quit);
  /*signal(SIGTRAP, sig_fatal); */
  signal (SIGIOT, sig_fatal);
  signal (SIGCHLD, sig_child);
  signal (SIGPIPE, SIG_IGN);

  if (!foreground)
    {
      int i;
      pid_t pid;

      pid = fork ();
      if (pid < 0)
	die (1, "cannot fork: %s", strerror (errno));
      if (pid > 0)
	{
	  /* Parent branch */
	  FILE *fp;
	  /* write pid file */
	  if ((fp = fopen (pidfile, "w")) != NULL)
	    {
	      fprintf (fp, "%d\n", pid);
	      fclose (fp);
	    }
	  else
	    {
	      logmsg (L_ERR, "can't write pid file %s: %s",
		      pidfile, strerror (errno));
	    }
	  exit (0);
	}
      /* Child branch */
      chdir ("/");
      for (i = 0; i < 3; i++)
	close (i);
      setsid ();
      use_stderr = 0;
    }

  change_user ();

  sin = (struct sockaddr_in *) &saremote;
  for (;;)
    {
      FD_ZERO (&read_fds);
      FD_SET (sockfd, &read_fds);

      rc = select (sockfd + 1, &read_fds, NULL, NULL, NULL);
      if (rc < 0)
	{
	  if (errno != EINTR)
	    logmsg (L_ERR, "select: %s", strerror (errno));
	  continue;
	}
      if (rc == 1)
	{
	  rc = recvfrom (sockfd, (char *) recv_buffer,
			 sizeof (i_recv_buffer), 0, &saremote, &salen);
	  if (rc > 0)
	    decode_buffer ();
	  else if (rc < 0 && errno != EINTR)
	    logmsg (L_ERR, "recvfrom: %s", strerror (errno));
	}
    }

}

void
decode_buffer ()
{
  int i;
  Stat_reply *reply;
  Stat *sp;
  int child = 0;

  reply = (Stat_reply *) recv_buffer;
  reply->n_addr = ntohl (reply->n_addr);

  if (reply->n_addr > MAXADDR)
    {
      logmsg (L_NOTICE, "got invalid packet: n_addr = %d", reply->n_addr);
      return;
    }
  reply->timestamp = ntohl (reply->timestamp);
  sp = reply->stat;

  if (!single_process_option)
    {
      pid_t pid = fork ();
      if (pid > 0)
	return;
      else if (pid < 0)
	logmsg (L_ERR, "can't fork: %s", strerror (errno));
      else
	{
	  signal (SIGHUP, SIG_IGN);
	  signal (SIGCHLD, SIG_IGN);
	  child = 1;
	}
    }

  open_db ();
  for (i = 0; i < reply->n_addr; i++, sp++)
    {
      sp->in = ntohl (sp->in);
      sp->out = ntohl (sp->out);
      logmsg (L_NOTICE, "Received %s %lu %lu", sp->name, sp->in, sp->out);
      report (sp);
    }
  close_db ();
  
  if (child)
    exit (0);
}

int
get_port (char *str)
{
  int pn;

  if (isdigit (str[0]))
    pn = htons (atoi (str));
  else
    {
      struct servent *s = getservbyname (str, "udp");
      if (s)
	pn = s->s_port;
      else
	{
	  logmsg (L_ERR, "no such service: %s", str);
	  return 0;
	}
    }
  return pn;
}

RETSIGTYPE
sig_quit (int sig)
{
  logmsg (L_INFO, "exiting on signal %d", sig);
  unlink (pidfile);
  exit (0);
}

RETSIGTYPE
sig_fatal (int sig)
{
  logmsg (L_ERR, "FATAL: exiting on signal %d", sig);
  unlink (pidfile);
  exit (1);
}

RETSIGTYPE
sig_hup (int sig)
{
  signal (sig, sig_hup);
}

RETSIGTYPE
sig_child (int sig)
{
  int status;

  waitpid ((pid_t) - 1, &status, WNOHANG);
  signal (sig, sig_child);
}

void
assign_string (char **pstr, const char *s)
{
  free (*pstr);
  if (s)
    {
      *pstr = xmalloc (strlen (s) + 1);
      strcpy (*pstr, s);
    }
  else
    *pstr = NULL;
}

void
assign_string_n (char **pstr, const char *s, size_t length)
{
  free (*pstr);
  if (s)
    {
      *pstr = xmalloc (length + 1);
      memcpy (*pstr, s, length);
      (*pstr)[length] = 0;
    }
  else
    *pstr = NULL;
}

char * 
mkfilename (char *dir, char *name, char *suffix)
{
  int ret;
  int len = strlen (name) + strlen (suffix);
  char *buf;
  if (dir)
    len += strlen (dir) + 1;
  buf = malloc (len + 1);
  if (buf)
    {
      buf[0] = 0;
      if (dir)
	{
	  strcat (buf, dir);
	  strcat (buf, "/");
	}
      strcat (buf, name);
      strcat (buf, suffix);
    }
  return buf;
}
